/*
 * Copyright (C) 2009-2014 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file wma_afh.c The audio format handler for WMA files. */

#include <stdlib.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <regex.h>

#include "para.h"
#include "error.h"
#include "afh.h"
#include "portable_io.h"
#include "string.h"
#include "wma.h"

#define FOR_EACH_FRAME(_f, _buf, _size, _ba) for (_f = (_buf); \
	_f + (_ba) + WMA_FRAME_SKIP < (_buf) + (_size); \
	_f += (_ba) + WMA_FRAME_SKIP)

/*
 * Must be called on a frame boundary, e.g. start + header_len.
 * \return Frame count, superframe count via *num_superframes.
 */
static int count_frames(const char *buf, int buf_size, int block_align,
	int *num_superframes)
{
	int fc = 0, sfc = 0; /* frame count, superframe count */
	const uint8_t *p;


	FOR_EACH_FRAME(p, (uint8_t *)buf, buf_size, block_align) {
		fc += p[WMA_FRAME_SKIP] & 0x0f;
		sfc++;
	}
	PARA_INFO_LOG("%d frames, %d superframes\n", fc, sfc);
	if (num_superframes)
		*num_superframes = sfc;
	return fc;
}

/*
 * put_utf8() and get_str16() below are based on macros in libavutil/common.h
 * of the mplayer source code, copyright (c) 2006 Michael Niedermayer
 * <michaelni@gmx.at>.
 */

/*
 * Convert a 32-bit Unicode character to its UTF-8 encoded form.
 *
 * Writes up to 4 bytes for values in the valid UTF-8 range and up to 7 bytes
 * in the general case, depending on the length of the converted Unicode
 * character.
 *
 * \param result Where the converted UTF-8 bytes are written.
 */
static int put_utf8(uint32_t val, char *result)
{
	char *out = result;
	int bytes, shift;
	uint32_t in = val;

	if (in < 0x80) {
		*out++ = in;
		return 1;
	}
	bytes = (wma_log2(in) + 4) / 5;
	shift = (bytes - 1) * 6;
	*out++ = (256 - (256 >> bytes)) | (in >> shift);
	while (shift >= 6) {
		shift -= 6;
		*out++ = 0x80 | ((in >> shift) & 0x3f);
	}
	return out - result;
}

static char *get_str16(const char *in, int len)
{
	const char *p = in;
	int out_size = 0, out_len = 0;
	char *out = NULL;

	len /= 2;
	while (len--) {
		uint32_t x;
		if (out_len + 7 + 1 >= out_size) {
			out_size = 2 * out_size + 50;
			out = para_realloc(out, out_size);
		}
		x = read_u16(p);
		p += 2;
		out_len += put_utf8(x, out + out_len);
		if (x == 0)
			return out;
	}
	if (out)
		out[out_len] = '\0';
	return out;
}

static const char comment_header[] = {
	0x33, 0x26, 0xb2, 0x75, 0x8E, 0x66, 0xCF, 0x11,
	0xa6, 0xd9, 0x00, 0xaa, 0x00, 0x62, 0xce, 0x6c
};

static const char extended_content_header[] = {
	0x40, 0xA4, 0xD0, 0xD2, 0x07, 0xE3, 0xD2, 0x11,
	0x97, 0xF0, 0x00, 0xA0, 0xC9, 0x5E, 0xA8, 0x50
};

static const char year_tag_header[] = { /* WM/Year */
	0x57, 0x00, 0x4d, 0x00, 0x2f, 0x00, 0x59, 0x00,
	0x65, 0x00, 0x61, 0x00, 0x72, 0x00
};

static const char album_tag_header[] = { /* WM/AlbumTitle */
	0x57, 0x00, 0x4d, 0x00, 0x2f, 0x00, 0x41, 0x00,
	0x6c, 0x00, 0x62, 0x00, 0x75, 0x00, 0x6d, 0x00,
	0x54, 0x00, 0x69, 0x00, 0x74, 0x00, 0x6c, 0x00,
	0x65, 0x00
};

static void read_asf_tags(const char *buf, int buf_size, struct taginfo *ti)
{
	const char *p, *end = buf + buf_size, *q;
	uint16_t len1, len2, len3, len4;

	p = search_pattern(comment_header, sizeof(comment_header),
		buf, buf_size);
	if (!p || p + 34 >= end) {
		PARA_NOTICE_LOG("comment header not found\n");
		goto next;
	}
	p += 24;
	len1 = read_u16(p);
	p += 2;
	len2 = read_u16(p);
	p += 2;
	len3 = read_u16(p);
	p += 2;
	len4 = read_u16(p);
	p += 2;
	/* ignore length of the rating information */
	p += 2;
	if (p + len1 >= end)
		goto next;
	ti->title = get_str16(p, len1);
	p += len1;
	if (p + len2 >= end)
		goto next;
	ti->artist = get_str16(p, len2);
	p += len2 + len3;
	if (p + len4 >= end)
		goto next;
	ti->comment = get_str16(p, len4);
next:
	p = search_pattern(extended_content_header, sizeof(extended_content_header),
		buf, buf_size);
	if (!p) {
		PARA_NOTICE_LOG("extended content header not found\n");
		return;
	}
	q = search_pattern(year_tag_header, sizeof(year_tag_header),
		p, end - p);
	if (q) {
		const char *r = q + sizeof(year_tag_header) + 6;
		if (r < end)
			ti->year = get_str16(r, end - r);
	}
	q = search_pattern(album_tag_header, sizeof(album_tag_header),
		p, end - p);
	if (q) {
		const char *r = q + sizeof(album_tag_header) + 6;
		if (r < end)
			ti->album = get_str16(r, end - r);
	}
}

static void set_chunk_tv(int frames_per_chunk, int frequency,
		struct timeval *result)
{
	uint64_t x = (uint64_t)frames_per_chunk * 2048 * 1000 * 1000
		/ frequency;

	result->tv_sec = x / 1000 / 1000;
	result->tv_usec = x % (1000 * 1000);
	PARA_INFO_LOG("chunk time: %lums\n", tv2ms(result));
}

/* Must be called on a frame boundary. */
static int wma_make_chunk_table(char *buf, size_t buf_size, int block_align,
		struct afh_info *afhi)
{
	const uint8_t *f, *start = (uint8_t *)buf;
	int j, frames_per_chunk;
	size_t ct_size = 250;
	int ret, count = 0, num_frames, num_superframes;

	afhi->chunk_table = para_malloc(ct_size * sizeof(uint32_t));
	afhi->chunk_table[0] = 0;
	afhi->chunk_table[1] = afhi->header_len;

	num_frames = count_frames(buf, buf_size, block_align,
		&num_superframes);
	ret = -E_NO_WMA;
	if (num_frames == 0 || num_superframes == 0)
		goto fail;
	afhi->seconds_total = num_frames * 2048 /* FIXME */
		/ afhi->frequency;
	frames_per_chunk = num_frames / num_superframes / 2;
	PARA_INFO_LOG("%d frames per chunk\n", frames_per_chunk);
	j = 1;
	FOR_EACH_FRAME(f, start, buf_size, block_align) {
		count += f[WMA_FRAME_SKIP] & 0x0f;
		while (count > j * frames_per_chunk) {
			j++;
			if (j >= ct_size) {
				ct_size *= 2;
				afhi->chunk_table = para_realloc(
					afhi->chunk_table,
					ct_size * sizeof(uint32_t));
			}
			afhi->chunk_table[j] = f - start + afhi->header_len + block_align + WMA_FRAME_SKIP;
		}
	}
	afhi->chunks_total = j;
	set_chunk_tv(frames_per_chunk, afhi->frequency, &afhi->chunk_tv);
	return 1;
fail:
	free(afhi->chunk_table);
	return ret;
}

static int wma_get_file_info(char *map, size_t numbytes, __a_unused int fd,
	struct afh_info *afhi)
{
	int ret;
	struct asf_header_info ahi;

	ret = read_asf_header(map, numbytes, &ahi);
	if (ret < 0)
		return ret;
	if (ret == 0)
		return -E_NO_WMA;
	afhi->bitrate = ahi.bit_rate / 1000;
	if (ahi.sample_rate == 0)
		return -E_NO_WMA;
	afhi->frequency = ahi.sample_rate;
	afhi->channels = ahi.channels;
	afhi->header_len = ahi.header_len;

	afhi->techinfo = make_message("%s%s%s%s%s",
		ahi.use_exp_vlc? "exp vlc" : "",
		(ahi.use_bit_reservoir && ahi.use_exp_vlc)? ", " : "",
		ahi.use_bit_reservoir? "bit reservoir" : "",
		(ahi.use_variable_block_len &&
			(ahi.use_exp_vlc || ahi.use_bit_reservoir)? ", " : ""),
		ahi.use_variable_block_len? "vbl" : ""
	);
	wma_make_chunk_table(map + ahi.header_len, numbytes - ahi.header_len,
		ahi.block_align, afhi);
	read_asf_tags(map, ahi.header_len, &afhi->tags);
	return 0;
}

static const char* wma_suffixes[] = {"wma", NULL};

/**
 * The init function of the wma audio format handler.
 *
 * \param afh Pointer to the struct to initialize.
 */
void wma_afh_init(struct audio_format_handler *afh)
{
	afh->get_file_info = wma_get_file_info;
	afh->suffixes = wma_suffixes;
}
